use std::io::{self, Write};
use std::path::{Path, PathBuf};
use std::sync::Arc;
use tokio::sync::Mutex;
use tokio::sync::{mpsc, oneshot};

use libarchive::write::{Builder, FileWriter, WriteEntry};
use libarchive::{Entry, WriteFilter, WriteFormat};

enum Message {
    AppendFilesEntry(oneshot::Sender<io::Result<()>>, String),
    AppendLine(oneshot::Sender<io::Result<()>>, String),
    Close(oneshot::Sender<io::Result<()>>),
}

/// Struct to abstract away the intrinsics of writing entries to an archive file
pub struct RepoArchiveWriter {
    tx: mpsc::Sender<Message>,
}

fn archive_manager_task(mut rx: mpsc::Receiver<Message>, mut ar: FileWriter) {
    // Once the accompanying struct and the final sender get dropped, this will return None and the
    // task will exit
    while let Some(msg) = rx.blocking_recv() {
        match msg {
            Message::AppendFilesEntry(tx, full_name) => {
                let mut ar_entry = WriteEntry::new();
                ar_entry.set_filetype(libarchive::archive::FileType::RegularFile);
                ar_entry.set_pathname(PathBuf::from(full_name).join("files"));
                ar_entry.set_mode(0o100644);
                // TODO set entry size?

                let mut res = ar.append_entry(&mut ar_entry).map_err(io::Error::from);

                if res.is_ok() {
                    res = writeln!(ar, "%FILES%");
                }

                tx.send(res);

                // All "files" entries start with this line
            }
            Message::AppendLine(tx, line) => {
                let res = writeln!(ar, "{}", line);

                tx.send(res);
            }
            Message::Close(tx) => {
                let res = ar.close();

                tx.send(res.map_err(io::Error::from));
            }
        }
    }
}

impl RepoArchiveWriter {
    pub async fn open<P: AsRef<Path>>(path: P) -> io::Result<Self> {
        let path = PathBuf::from(path.as_ref());
        let (tx, rx) = mpsc::channel(1);

        // Open the archive file
        let ar = tokio::task::spawn_blocking(move || {
            let mut builder = Builder::new();
            builder.add_filter(WriteFilter::Gzip)?;
            builder.set_format(WriteFormat::PaxRestricted)?;

            builder.open_file(path)
        })
        .await
        .unwrap()?;

        // Spawn blocking task to perform blocking actions
        tokio::task::spawn_blocking(move || archive_manager_task(rx, ar));

        Ok(Self { tx })
    }

    /// Set the current entry to be a new "files" list
    pub async fn add_files_entry(&self, full_name: &str) -> io::Result<()> {
        let full_name = String::from(full_name);
        let (tx, rx) = oneshot::channel();

        self.tx.send(Message::AppendFilesEntry(tx, full_name)).await;

        rx.await.unwrap()
    }
    //
    ///// Append the given line to the currently active entry
    pub async fn write_line(&self, line: &str) -> io::Result<()> {
        let line = String::from(line);
        let (tx, rx) = oneshot::channel();

        self.tx.send(Message::AppendLine(tx, line)).await;

        rx.await.unwrap()
    }

    pub async fn close(&self) -> io::Result<()> {
        let (tx, rx) = oneshot::channel();

        self.tx.send(Message::Close(tx)).await;

        rx.await.unwrap()
    }
}
